/*
 * Bayer-mask pixel processing core for BGBG/GRGR arrays (as seen in the
 * Omnivision OV2610 and OV3610 chips and many more).
 *
 * Copyright (c) 2004 Steve Hosgood <steve@caederus.com> Jul-Nov 2004
 *
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */
 
//# define INTELLIGENT_BAYER		/* better colour processing (allegedly). Increases processor load by 1.5 */			

# include	<linux/config.h>
# include	<linux/version.h>
# include	<linux/module.h>
# include	<linux/kernel.h>
# include	<linux/init.h>		/* don't really need this post 2.6.5 */


/*
 * Version Information
 */
# define DESC		"BGBG/GRGR Bayer mask processor"
# define VERSION	"v0.99"
# define EMAIL		"steve@caederus.com"
# define AUTHOR		"Steve Hosgood <steve@caederus.com>"
# define COPYRIGHT	"1994, XL Imaging Ltd, Swansea SA2 8PP, United Kingdom."

//# define BAYER_DEBUG
# ifdef BAYER_DEBUG
static int debug;

MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug,
  "Debug level: 0=none, 1=inits, 2=warning, 3=config, 4=functions, 5=max");
# endif

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESC);
#if defined(MODULE_LICENSE)	/* Introduced in ~2.4.10 */
MODULE_LICENSE("GPL");
#endif

# include	"bayer_bgbg_grgr.h"	/* self-check */

/*
 * Notes: this Bayer processing will only work if the output frame has the
 * same pixel dimensions as the raw frame.
 */
void
bayer_decode_BGBG_GRGR_8( unsigned char *pOut, unsigned char *pIn, int width, int height, int flags, matrix_f *rematrix_even, matrix_f *rematrix_odd )
{
	unsigned char	*pAbove = pIn, *pBelow = pIn + width;
	short		r, g, b;
# ifdef INTELLIGENT_BAYER				
	int		d1, d2;
# endif
	unsigned short	x, y;

	/* Top line (line 0) is an even line, i.e BGBG */
	/* Top left corner is B. Interpolate the other two */
	b = pIn[0];
	r = pBelow[1];
	g = (pIn[1] + pBelow[0]) >> 1;
	pOut = rematrix_even(pOut, r, g, b);
	pIn++;
	pBelow++;
	
	for (x = 1; x < width - 1; x += 2) {
		/* Rest of even line runs GBGB... */
		/* This pixel is G. Interpolate the other two */
		g = pIn[0];
		b = (pIn[-1] + pIn[1]) >> 1;
		r = pBelow[0];
		pOut = rematrix_odd(pOut, r, g, b);
		pIn++;
		pBelow++;

		/* This pixel is B. Interpolate the other two */
		b = pIn[0];
		g = pBelow[0];
		r = (pBelow[-1] + pBelow[1]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
		pIn++;
		pBelow++;
	}
	
	/* Top right corner is G. Interpolate the other two */
	g = pIn[0];
	r = pBelow[0];
	b = pIn[-1];
	pOut = rematrix_odd(pOut, r, g, b);
	pIn++;
	pBelow++;
	
	/* Now for all the main-body lines */
	/* 'pAbove' is useful now that we've processed the top (even) line */	
	for (y = 1; y < height - 1; y += 2) {
		/* Odd lines are GRGR */
		/* Left margin is G. Interpolate the other two */
		g = pIn[0];
		r = pIn[1];
		b = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
		pIn++;
		pBelow++;
		pAbove++;

		for (x = 1; x < width - 1; x += 2) {
			/* Rest of odd line runs RGRG... */
			/* This pixel is R. Interpolate the other two */
			r = pIn[0];
# ifdef INTELLIGENT_BAYER
			/* a choice of four for B: be intelligent */
			if ((d1 = pAbove[-1] - pBelow[1]) < 0)
				d1 = -d1;
			if ((d2 = pAbove[1] - pBelow[-1]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				b = (pAbove[-1] + pBelow[1]) >> 1;
			else if (d1 > d2)
				b = (pAbove[1] + pBelow[-1]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				b = (pAbove[-1] + pBelow[-1] + pAbove[1] + pBelow[1]) >> 2;
				
# ifdef INTELLIGENT_BAYER
			/* a choice of four for G: be intelligent */
			if ((d1 = pAbove[0] - pBelow[0]) < 0)
				d1 = -d1;
			if ((d2 = pIn[-1] - pIn[1]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				g = (pAbove[0] + pBelow[0]) >> 1;
			else if (d1 > d2)
				g = (pIn[-1] + pIn[1]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				g = (pAbove[0] + pBelow[0] + pIn[-1] + pIn[1]) >> 2;
				
			pOut = rematrix_odd(pOut, r, g, b);
			pIn++;
			pBelow++;
			pAbove++;

			/* This pixel is G. Interpolate the other two */
			g = pIn[0];
			b = (pAbove[0] + pBelow[0]) >> 1;
			r = (pIn[-1] + pIn[1]) >> 1;
			pOut = rematrix_even(pOut, r, g, b);
			pIn++;
			pBelow++;
			pAbove++;
		}
		
		/* Right margin is R. Interpolate the other two */
		r = pIn[0];
		b = (pAbove[-1] + pBelow[-1]) >> 1;
		g = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_odd(pOut, r, g, b);
		pIn++;
		pBelow++;
		pAbove++;

		/* Even lines are BGBG */
		/* Left margin is B. Interpolate the other two */
		b = pIn[0];
		r = (pAbove[1] + pBelow[1]) >> 1;
		g = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
		pIn++;
		pBelow++;
		pAbove++;

		for (x = 1; x < width - 1; x += 2) {
			/* Rest of even line runs GBGB... */
			/* This pixel is G. Interpolate the other two */
			g = pIn[0];
			r = (pAbove[0] + pBelow[0]) >> 1;
			b = (pIn[-1] + pIn[1]) >> 1;
			pOut = rematrix_odd(pOut, r, g, b);
			pIn++;
			pBelow++;
			pAbove++;
			
			/* This pixel is B. Interpolate the other two */
			b = pIn[0];
			
# ifdef INTELLIGENT_BAYER
			/* a choice of four for R: be intelligent */
			if ((d1 = pAbove[-1] - pBelow[1]) < 0)
				d1 = -d1;
			if ((d2 = pAbove[1] - pBelow[-1]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				r = (pAbove[-1] + pBelow[1]) >> 1;
			else if (d1 > d2)
				r = (pAbove[1] + pBelow[-1]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				r = (pAbove[-1] + pBelow[-1] + pAbove[1] + pBelow[1]) >> 2;
				
# ifdef INTELLIGENT_BAYER
			/* a choice of four for G: be intelligent */
			if ((d1 = pAbove[0] - pBelow[0]) < 0)
				d1 = -d1;
			if ((d2 = pIn[-1] - pIn[1]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				g = (pAbove[0] + pBelow[0]) >> 1;
			else if (d1 > d2)
				g = (pIn[-1] + pIn[1]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				g = (pAbove[0] + pBelow[0] + pIn[-1] + pIn[1]) >> 2;
				
			pOut = rematrix_even(pOut, r, g, b);
			pIn++;
			pBelow++;
			pAbove++;
		}
		
		/* Right margin is G. Interpolate the other two */
		g = pIn[0];
		b = pIn[-1];
		r = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_odd(pOut, r, g, b);
		pIn++;
		pBelow++;
		pAbove++;
	}

	/* Bottom line (line 1199) is an odd line, i.e GRGR */
	/* Bottom left corner is G. Interpolate the other two */
	g = pIn[0];
	r = pIn[1];
	b = pAbove[0];
	pOut = rematrix_even(pOut, r, g, b);
	pIn++;
	pAbove++;
	
	for (x = 1; x < width - 1; x += 2) {
		/* Rest of odd line runs RGRG... */
		/* This pixel is R. Interpolate the other two */
		r = pIn[0];
		b = (pAbove[-1] + pAbove[1]) >> 1;
		g = pAbove[0];
		pOut = rematrix_odd(pOut, r, g, b);
		pIn++;
		pAbove++;

		/* This pixel is G. Interpolate the other two */
		g = pIn[0];
		b = pAbove[0];
		r = (pIn[-1] + pIn[1]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
		pIn++;
		pAbove++;
	}
	
	/* Bottom right corner is R. Interpolate the other two */
	r = pIn[0];
	g = (pAbove[0] + pIn[-1]) >> 1;
	b = pAbove[-1];
	rematrix_odd(pOut, r, g, b);
	return;
}


# ifdef DOUBLEUP
# error "Doubleup() dates from the days of the Old Testament: you'll have to"
# error "rework it for sure."
static unsigned char *
doubleup( unsigned char *pIn, int format )
{

	if (format == VIDEO_PALETTE_RGB24) {
		*pIn = pIn[-3];
		pIn++;
		*pIn = pIn[-3];
		pIn++;
		*pIn = pIn[-3];
		pIn++;
	}
	else if (format == VIDEO_PALETTE_RGB32) {
		*pIn = pIn[-4];
		pIn++;
		*pIn = pIn[-4];
		pIn++;
		*pIn = pIn[-4];
		pIn++;
		*pIn = pIn[-4];
		pIn++;
	}
	else {
		/* a 16-bit format */
		*pIn = pIn[-2];
		pIn++;
		*pIn = pIn[-2];
		pIn++;
	}
		
	return pIn;
}
# endif /* DOUBLEUP */


/*
 * Notes: this Bayer processing will only work if the output frame has the same
 * pixel dimensions as the raw frame.
 *
 * Additionally, the raw frame is organised in words, not bytes.
 *
 * # ifdef DOUBLEUP
 *	If ov->scaling is set, we resample the frame to 1600x1200 even though the user asked for 800x600.
 *	This is a mighty hack.
 * # endif
 */
void
bayer_decode_BGBG_GRGR_16( unsigned char *pOut, unsigned char *pIn, int width, int height, int flags, matrix_f *rematrix_even, matrix_f *rematrix_odd )
{
	unsigned char	*pAbove = pIn, *pBelow = pIn + width*2;
	short		r, g, b;
# ifdef INTELLIGENT_BAYER				
	int		d1, d2;
# endif
	unsigned short	x, y;
	
# ifdef DOUBLEUP
	if (width != 800) {
		PDEBUG(MISC|4, "Frame was not 800x600");
		return;
	}
# endif /* DOUBLEUP */

	/* Top line (line 0) is an even line, i.e BGBG */
	/* Top left corner is B. Interpolate the other two */
	b = pIn[0];
	r = pBelow[2];
	g = (pIn[2] + pBelow[0]) >> 1;
	pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
	pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
	pIn += 2;
	pBelow += 2;
	
	for (x = 1; x < width - 1; x += 2) {
		/* Rest of even line runs GBGB... */
		/* This pixel is G. Interpolate the other two */
		g = pIn[0];
		b = (pIn[-2] + pIn[2]) >> 1;
		r = pBelow[0];
		pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;

		pBelow += 2;

		/* This pixel is B. Interpolate the other two */
		b = pIn[0];
		g = pBelow[0];
		r = (pBelow[-2] + pBelow[2]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pBelow += 2;
	}
	
	/* Top right corner is G. Interpolate the other two */
	g = pIn[0];
	r = pBelow[0];
	b = pIn[-2];
	pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
	pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
	pIn += 2;
	pBelow += 2;

# ifdef DOUBLEUP
	if (format == VIDEO_PALETTE_RGB24) {	
		memcpy(pOut, &pOut[-3*1600], 3*1600);
		pOut += 3*1600;
	}
	else if (format == VIDEO_PALETTE_RGB32) {	
		memcpy(pOut, &pOut[-4*1600], 4*1600);
		pOut += 4*1600;
	}
	else {	
		memcpy(pOut, &pOut[-2*1600], 2*1600);
		pOut += 2*1600;
	}
# endif /* DOUBLEUP */
	
	/* Now for all the main-body lines */
	/* 'pAbove' is useful now that we've processed the top (even) line */	
	for (y = 1; y < height - 1; y += 2) {
		/* Odd lines are GRGR */
		/* Left margin is G. Interpolate the other two */
		g = pIn[0];
		r = pIn[2];
		b = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pBelow += 2;
		pAbove += 2;

		for (x = 1; x < width - 1; x += 2) {
			/* Rest of odd line runs RGRG... */
			/* This pixel is R. Interpolate the other two */
			r = pIn[0];
# ifdef INTELLIGENT_BAYER
			/* a choice of four for B: be intelligent */
			if ((d1 = pAbove[-2] - pBelow[2]) < 0)
				d1 = -d1;
			if ((d2 = pAbove[2] - pBelow[-2]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				b = (pAbove[-2] + pBelow[2]) >> 1;
			else if (d1 > d2)
				b = (pAbove[2] + pBelow[-2]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				b = (pAbove[-2] + pBelow[-2] + pAbove[2] + pBelow[2]) >> 2;
				
# ifdef INTELLIGENT_BAYER
			/* a choice of four for G: be intelligent */
			if ((d1 = pAbove[0] - pBelow[0]) < 0)
				d1 = -d1;
			if ((d2 = pIn[-2] - pIn[2]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				g = (pAbove[0] + pBelow[0]) >> 1;
			else if (d1 > d2)
				g = (pIn[-2] + pIn[2]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				g = (pAbove[0] + pBelow[0] + pIn[-2] + pIn[2]) >> 2;
				
			pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
			pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
			pIn += 2;
			pBelow += 2;
			pAbove += 2;

			/* This pixel is G. Interpolate the other two */
			g = pIn[0];
			b = (pAbove[0] + pBelow[0]) >> 1;
			r = (pIn[-2] + pIn[2]) >> 1;
			pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
			pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
			pIn += 2;
			pBelow += 2;
			pAbove += 2;
		}
		
		/* Right margin is R. Interpolate the other two */
		r = pIn[0];
		b = (pAbove[-2] + pBelow[-2]) >> 1;
		g = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pBelow += 2;
		pAbove += 2;
		
# ifdef DOUBLEUP
		if (format == VIDEO_PALETTE_RGB24) {	
			memcpy(pOut, &pOut[-3*1600], 3*1600);
			pOut += 3*1600;
		}
		else if (format == VIDEO_PALETTE_RGB32) {	
			memcpy(pOut, &pOut[-4*1600], 4*1600);
			pOut += 4*1600;
		}
		else {	
			memcpy(pOut, &pOut[-2*1600], 2*1600);
			pOut += 2*1600;
		}
# endif /* DOUBLEUP */

		/* Even lines are BGBG */
		/* Left margin is B. Interpolate the other two */
		b = pIn[0];
		r = (pAbove[2] + pBelow[2]) >> 1;
		g = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pBelow += 2;
		pAbove += 2;

		for (x = 1; x < width - 1; x += 2) {
			/* Rest of even line runs GBGB... */
			/* This pixel is G. Interpolate the other two */
			g = pIn[0];
			r = (pAbove[0] + pBelow[0]) >> 1;
			b = (pIn[-2] + pIn[2]) >> 1;
			pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
			pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
			pIn += 2;
			pBelow += 2;
			pAbove += 2;
			
			/* This pixel is B. Interpolate the other two */
			b = pIn[0];
			
# ifdef INTELLIGENT_BAYER
			/* a choice of four for R: be intelligent */
			if ((d1 = pAbove[-2] - pBelow[2]) < 0)
				d1 = -d1;
			if ((d2 = pAbove[2] - pBelow[-2]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				r = (pAbove[-2] + pBelow[2]) >> 1;
			else if (d1 > d2)
				r = (pAbove[2] + pBelow[-2]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				r = (pAbove[-2] + pBelow[-2] + pAbove[2] + pBelow[2]) >> 2;
				
# ifdef INTELLIGENT_BAYER
			/* a choice of four for G: be intelligent */
			if ((d1 = pAbove[0] - pBelow[0]) < 0)
				d1 = -d1;
			if ((d2 = pIn[-2] - pIn[2]) < 0)
				d2 = -d2;
				
			if (d1 < d2)
				g = (pAbove[0] + pBelow[0]) >> 1;
			else if (d1 > d2)
				g = (pIn[-2] + pIn[2]) >> 1;
			else
# endif /* INTELLIGENT_BAYER */
				g = (pAbove[0] + pBelow[0] + pIn[-2] + pIn[2]) >> 2;
				
			pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
			pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
			pIn += 2;
			pBelow += 2;
			pAbove += 2;
		}
		
		/* Right margin is G. Interpolate the other two */
		g = pIn[0];
		b = pIn[-2];
		r = (pAbove[0] + pBelow[0]) >> 1;
		pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pBelow += 2;
		pAbove += 2;
		
# ifdef DOUBLEUP
		if (format == VIDEO_PALETTE_RGB24) {	
			memcpy(pOut, &pOut[-3*1600], 3*1600);
			pOut += 3*1600;
		}
		else if (format == VIDEO_PALETTE_RGB32) {	
			memcpy(pOut, &pOut[-4*1600], 4*1600);
			pOut += 4*1600;
		}
		else {	
			memcpy(pOut, &pOut[-2*1600], 2*1600);
			pOut += 2*1600;
		}
# endif /* DOUBLEUP */
	}

	/* Bottom line (line 1199) is an odd line, i.e GRGR */
	/* Bottom left corner is G. Interpolate the other two */
	g = pIn[0];
	r = pIn[2];
	b = pAbove[0];
	pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
	pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
	pIn += 2;
	pAbove += 2;
	
	for (x = 1; x < width - 1; x += 2) {
		/* Rest of odd line runs RGRG... */
		/* This pixel is R. Interpolate the other two */
		r = pIn[0];
		b = (pAbove[-2] + pAbove[2]) >> 1;
		g = pAbove[0];
		pOut = rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pAbove += 2;

		/* This pixel is G. Interpolate the other two */
		g = pIn[0];
		b = pAbove[0];
		r = (pIn[-2] + pIn[2]) >> 1;
		pOut = rematrix_even(pOut, r, g, b);
# ifdef DOUBLEUP
		pOut = doubleup(pOut, format);
# endif /* DOUBLEUP */
		pIn += 2;
		pAbove += 2;
	}
	
	/* Bottom right corner is R. Interpolate the other two */
	r = pIn[0];
	g = (pAbove[0] + pIn[-2]) >> 1;
	b = pAbove[-2];
	rematrix_odd(pOut, r, g, b);
# ifdef DOUBLEUP
	pOut = doubleup(pOut, format);
	
	if (format == VIDEO_PALETTE_RGB24) {	
		memcpy(pOut, &pOut[-3*1600], 3*1600);
		pOut += 3*1600;
	}
	else if (format == VIDEO_PALETTE_RGB32) {	
		memcpy(pOut, &pOut[-4*1600], 4*1600);
		pOut += 4*1600;
	}
	else {	
		memcpy(pOut, &pOut[-2*1600], 2*1600);
		pOut += 2*1600;
	}
# endif /* DOUBLEUP */
	return;
}


/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

/*
 * We come here as a consequence of loading the driver with '/sbin/insmod'
 */ 
static int __init
init(void)
{

	/* NB: string pasting below! */
	printk(KERN_INFO "Loaded " DESC " " VERSION "\n");
	printk(KERN_INFO "Written by: " AUTHOR "\n");
	printk(KERN_INFO "(c) " COPYRIGHT "\n");
	return 0;
}


/*
 * We come here as a consequence of unloading the driver with '/sbin/rmmod'
 */ 
static void __exit
cleanup(void)
{

	/* NB: string pasting below! */
	printk(KERN_INFO DESC " unloaded\n");
	return;
}


# ifdef MODULE
module_init(init);
module_exit(cleanup);
# endif
