/* Copyright (c) 1999-2003 Mark W. McClelland, David Brownell
 * http://alpha.dyndns.org/ov511/
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */

#ifndef __LINUX_OVFX2_H
#define __LINUX_OVFX2_H

#include <linux/config.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/i2c.h>

#include "ovcamchip.h"
#include "bayer.h"


#ifdef MODULE_DEBUG
#define FNCSHIFT	4	/* best if this is 4,8,12 etc */
#define LMASK		((1 << FNCSHIFT) - 1)
#define MISC		(0x01 << FNCSHIFT)
#define FRAMESTATE	(0x02 << FNCSHIFT)
#define INCOMING	(0x04 << FNCSHIFT)
#define OUTGOING	(0x08 << FNCSHIFT)
#define IOCTL		(0x10 << FNCSHIFT)	/* debug=0x10x */
#define FNCS		(0x20 << FNCSHIFT)	/* debug=0x20x */
#define BOMB		(0x40 << FNCSHIFT)	/* debug=0x40x */

	#define PDEBUG(level, fmt, args...) \
		if ((debug & (level) & ~LMASK) != 0 && (debug & LMASK) >= ((level) & LMASK)) \
			info("[%s:%d] " fmt, __PRETTY_FUNCTION__, __LINE__ , ## args)
#else
	#define PDEBUG(level, fmt, args...) do {} while(0)
#endif

/* This macro restricts an int variable to an inclusive range */
#define RESTRICT_TO_RANGE(v,mi,ma) { \
	if ((v) < (mi)) (v) = (mi); \
	else if ((v) > (ma)) (v) = (ma); \
}

/* --------------------------------- */
/* DEFINES FOR OV511 AND OTHER CHIPS */
/* --------------------------------- */

/* USB IDs */
#define VEND_OMNIVISION		0x05a9
#define VEND_ORANGE_MICRO	0x0b62
#define VEND_CYPRESS		0x04b4
#define VEND_XLI		0x15C0

#define PROD_IBOT2		0x0059
#define PROD_TEST		0x2800
#define PROD_XLI_2M		0x0001

/* --------------------------------- */
/*      FX2 REGISTER MNEMONICS       */
/* --------------------------------- */

#define REG_I2C_ADDR		0x00
#define REG_SAT			0xe9
#define REG_HUE			0xea
#define REG_CNTR		0xeb
#define REG_BRIGHT		0xec
#define REG_SHARP		0xee

/* --------------------------------- */
/*       MISCELLANEOUS DEFINES       */
/* --------------------------------- */

/* USB bRequest values */
#define OVFX2_REQ_I2C_WRITE	0x02
#define OVFX2_REQ_I2C_READ	0x03
#define OVFX2_REQ_REG_WRITE	0x0a
#define OVFX2_REQ_REG_READ	0x0b

/* Endpoints */
#define OVFX2_INT_ENDPOINT_ADDR		1
#define OVFX2_BULK_ENDPOINT_ADDR	2

/* # and size of BULK scratch buffers */
#define OVFX2_NUMSBUF		16 /* was 8 */
#define OVFX2_BULK_SIZE		(16*1024) /* was 32768 */

#define OVFX2_NUMFRAMES		2	/* don't change this without fixing FIXME */
#if OV511_NUMFRAMES > VIDEO_MAX_FRAME
	#error "OV511_NUMFRAMES is too high"
#endif

/* All known registers are only 1 byte */
#define OVFX2_CBUF_SIZE		1

/* Bridge types */
enum {
	BRG_UNKNOWN,
	BRG_2800,	/* iBot2, OV "2800" ref design */
	BRG_9000,	/* OV "9000" ref design */
	BRG_2610,	/* OV2610 ref design */
};

/* Buffer states */
enum {
	BUF_NOT_ALLOCATED,
	BUF_ALLOCATED,
};

struct usb_ovfx2;		/* Forward declaration */

struct ovfx2_sbuf {
	struct usb_ovfx2 *ov;
	unsigned char *data;
	struct urb *urb;
	spinlock_t lock;
	int n;
};

/* state machine */
enum state_machine {
	FREE_EMPTY,		/* Empty (no MCAPTURE) */
	MARKED_EMPTY,		/* Empty (MCAPTURE issued) */
	FREE_FILLING,		/* Filling with bulk data (no MCAPTURE) */
	MARKED_FILLING,		/* Filling with bulk data (MCAPTURE issued) */
	FREE_FULL,		/* Filled with bulk data (no MCAPTURE) */
	MARKED_FULL,		/* Filled with bulk data, (MCAPTURE issued, but not been synced yet) */
	FRAME_ERROR,		/* Something bad happened while processing */
	FRAME_ABORT,		/* bomb any CSYNC pending on this frame */
};

enum fx2_state {
	FX2_NOT_STREAMING,
	FX2_STREAMING,
	FX2_PAUSED,		/* we pause during sensor mode-switching */
	FX2_WEDGED,		/* the timebomb went off, we *were* streaming */
	FX2_WRONGMODE,		/* we want to capture a frame, but the FX2 needs a mode change */
	FX2_WRONGSIZE,		/* FX2 is giving us the wrong size frames */
};

enum sensor_state {
	SENSOR_UNKNOWN_MODE,	/* not yet initialised */
	SENSOR_SVGA,		/* 800x600 mode */
	SENSOR_UXGA,		/* 1200x1600 mode */
};

struct regval {
	unsigned char reg;
	unsigned char val;
	unsigned char mask;
};

struct ovfx2_frame {
	int framenum;		/* Index of this frame */
	unsigned char *data;	/* Frame buffer */
# ifdef TBUF
	unsigned char *tempdata; /* Temp buffer for multi-stage conversions */
# endif
	unsigned char *rawdata;	/* Raw camera data buffer */

	int depth;		/* Bytes per pixel */
	int width;		/* Width application is expecting */
	int height;		/* Height application is expecting */

	int rawwidth;		/* Actual width of frame sent from camera */
	int rawheight;		/* Actual height of frame sent from camera */

	int sub_flag;		/* Sub-capture mode for this frame? */
	matrix_f *rematrix_even; /* pixel-packer for this frame's format */
	matrix_f *rematrix_odd; /* pixel-packer for this frame's format */

	volatile enum state_machine state;	/* State machine to control frame-handling */

	int bytes_recvd;	/* Number of image bytes received from camera */
	long bytes_read;	/* Amount that has been read() */
	enum sensor_state reqdsensorstate; /* 800x600 or 1600x1200? */
	bayer_f *bayer_decode;

	wait_queue_head_t wq;	/* Processes waiting */
};

struct usb_ovfx2 {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_interface *intf;
#else
	struct usb_device *dev;
#endif
	int minor;	/* V4L minor device number */
	char *desc;

	/* Determined by sensor type */
	int maxwidth;
	int maxheight;
	int minwidth;
	int minheight;

	struct semaphore lock;	/* Serializes user-accessible operations */

	enum fx2_state fx2state;	/* Are we streaming BULK? */
	enum sensor_state sensorstate; /* 800x600 or 1600x1200? */
	unsigned int currentformat;	/* Format for this frame as set by VIDIOCMCAPTURE */
	int	fsync;			/* last bulk frame was undersized: next one starts a frame */
//	int	scaling;		/* 0: output 800x600 as 800x600, 1: double up to 1600x1200 */

	unsigned char *fbuf;	/* Videodev buffer area */
# ifdef TBUF
	unsigned char *tempfbuf; /* Temporary (intermediate) buffer area */
# endif
	unsigned char *rawfbuf;	/* Raw camera data buffer area. This gets deallocated if the device 'goes away' or no-one's got an open file using the device */

	int sub_flag;		/* Pix Array subcapture on flag */
	int subx;		/* Pix Array subcapture x offset */
	int suby;		/* Pix Array subcapture y offset */
	int subw;		/* Pix Array subcapture width */
	int subh;		/* Pix Array subcapture height */

	int bridge;		/* Type of bridge (BRG_*) */
	int sensor;		/* Type of image sensor chip (SEN_*) */

	/* /proc entries, relative to /proc/video/ov511/ */
	struct proc_dir_entry *proc_devdir;   /* Per-device proc directory */
	struct proc_dir_entry *proc_info;     /* <minor#>/info entry */

	/* Framebuffer/sbuf management */
	int buf_state;
	struct semaphore buf_lock;
	volatile int curframe;		/* Frame being written to */
	struct ovfx2_frame frame[OVFX2_NUMFRAMES];	
	struct ovfx2_sbuf sbuf[OVFX2_NUMSBUF];
	wait_queue_head_t wq;	/* Processes waiting */

	/* Stop streaming while changing picture settings */
	int stop_during_set;

	/* Internal I2C interface */
	struct i2c_client internal_client; /* Client for internal use */

	/* I2C interface to kernel */
	struct i2c_adapter i2c_adap;
	struct i2c_client *sensor_client;
	unsigned char last_slave;	   /* Last accessed I2C slave */

	/* Control transaction stuff */
	unsigned char *cbuf;		/* Buffer for payload */
	struct semaphore cbuf_lock;
	struct timer_list	timebomb;
	int			bomb_running;
};

/* Used to represent a list of values and their respective symbolic names */
struct symbolic_list {
	int num;
	char *name;
};

#define NOT_DEFINED_STR "Unknown"

/* Returns the name of the matching element in the symbolic_list array. The
 * end of the list must be marked with an element that has a NULL name.
 */
static inline char * 
symbolic(struct symbolic_list list[], int num)
{
	int i;

	for (i = 0; list[i].name != NULL; i++)
			if (list[i].num == num)
				return (list[i].name);

	return (NOT_DEFINED_STR);
}



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
extern struct usb_ovfx2	*ovfx2_probe( struct usb_interface *intf, const struct usb_device_id *id, int minor );
#else
extern struct usb_ovfx2	*ovfx2_probe( struct usb_device *dev, const struct usb_device_id *id, int minor );
#endif

extern int		ovfx2_initial_open( struct usb_ovfx2 *ov, struct file *file );
extern int		ovfx2_ioctl( struct usb_ovfx2 *ov, struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg );
extern int		ovfx2_read( struct usb_ovfx2 *ov, struct file *file, char *buf, size_t cnt, loff_t *ppos );
extern int		ovfx2_read( struct usb_ovfx2 *ov, struct file *file, char *buf, size_t cnt, loff_t *ppos );
extern int		ovfx2_mmap( struct usb_ovfx2 *ov, struct file *file, struct vm_area_struct *vma );
extern int		ovfx2_final_close( struct usb_ovfx2 *ov );

extern void		ovfx2_disconnect( struct usb_ovfx2 *ov );
extern void		ovfx2_free( struct usb_ovfx2 *ov );

# ifdef MODULE
EXPORT_SYMBOL(ovfx2_probe);
EXPORT_SYMBOL(ovfx2_initial_open);
EXPORT_SYMBOL(ovfx2_ioctl);
EXPORT_SYMBOL(ovfx2_read);
EXPORT_SYMBOL(ovfx2_mmap);
EXPORT_SYMBOL(ovfx2_final_close);
EXPORT_SYMBOL(ovfx2_disconnect);
EXPORT_SYMBOL(ovfx2_free);
# endif

#endif
